"""A Vulnerability Scanner that combines results from all configured scanners"""
# pylint: disable=duplicate-code
from pathlib import Path
from typing import List

import typer
from typer import Typer

from security_commons.common.reporting.models import ReportFormat
from security_commons.common.reporting.reporting import Reporting
from security_commons.common.utils import parse_sbom, parse_sbom_json_string
from hopprcop.combined.combined_scanner import CombinedScanner

app = Typer()


@app.command()
def vulnerability_report(
    bom: str = typer.Argument(None, help="the path to a cyclone-dx bom or json value"),
    formats: List[ReportFormat] = typer.Option(
        ["table"],
        "--format",
        help="The report formats to generate",
    ),
    output_dir: Path = typer.Option(
        Path.cwd(), help="The directory where reports will be writen"
    ),
    base_report_name: str = typer.Option(
        None, help="The base name supplied for the generated reports"
    ),
):
    """Runs hoppr cop from cli"""
    if base_report_name is None:
        if bom.endswith(".json"):
            base_report_name = bom.removesuffix(".json")
        elif bom.endswith(".xml"):
            base_report_name = bom.removesuffix(".xml")
        else:
            base_report_name = "hoppr-cop-report"

    run_hoppr_cop(bom, base_report_name, formats, get_scanners(), output_dir)


def get_scanners() -> List[str]:
    """Defines scanners to use for hoppr cop"""
    return [
        "hopprcop.gemnasium.gemnasium_scanner.GemnasiumScanner",
        "hopprcop.grype.grype_scanner.GrypeScanner",
        "hopprcop.ossindex.oss_index_scanner.OSSIndexScanner",
    ]


def run_hoppr_cop(
    bom: str,
    base_report_name: str,
    formats: List[ReportFormat],
    scanners: List[str],
    output_dir: Path = Path.cwd(),
):
    """generates vulnerability reports based on the specified BOM and formats"""
    reporting = Reporting(output_dir, base_report_name)
    combined = CombinedScanner()
    combined.set_scanners(scanners)

    parsed_bom = None

    if bom.endswith(".json") or bom.endswith(".xml"):
        parsed_bom = parse_sbom(Path(bom))
    else:
        parsed_bom = parse_sbom_json_string(bom, "The json provided sbom")

    result = combined.get_vulnerabilities_by_sbom(parsed_bom)
    reporting.generate_vulnerability_reports(formats, result, parsed_bom)



from pathlib import Path

from hoppr.base_plugins.hoppr import HopprPlugin, hoppr_process
from hoppr.result import Result
from hoppr.hoppr_types.bom_access import BomAccess
from security_commons.common.reporting.models import ReportFormat

from hopprcop.combined.cli import run_hoppr_cop, get_scanners
from hopprcop import __version__


class HopprCopPlugin(HopprPlugin):
    """
    hoppr plugin wrapper for hopprcop integration
    """

    bom_access = BomAccess.FULL_ACCESS

    def get_version(self) -> str:
        """
        __version__ required for all HopprPlugin implementations
        """
        return __version__

    @hoppr_process
    def pre_stage_process(self) -> Result:
        """
        Supply sbom to hoppr cop to perform vulnerabilty check
        """
        self.get_logger().info("[ Executing hopprcop vulnerability check ]")
        self.get_logger().flush()

        output_dir = self.config.get("output_dir", Path.cwd())
        base_report_name = self.config.get(
            "base_report_name", "hopprcop-vulnerability-results"
        )
        scanners = self.config.get("scanners", get_scanners())
        formats = self.config.get("result_formats", [ReportFormat.CYCLONE_DX])

        run_hoppr_cop(
            self.context.delivered_sbom.json(),
            base_report_name,
            formats,
            scanners,
            output_dir,
        )

        return Result.success()



"""A Vulnerability Scanner that combines results from all configured scanners"""
# This file is part of hoppr-cop
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Copyright (c) 2022 Lockheed Martin Corporation
import concurrent.futures
import importlib
from typing import List, Optional, Union

import typer

from packageurl import PackageURL

from hoppr_cyclonedx_models.cyclonedx_1_3 import (
    CyclonedxSoftwareBillOfMaterialSpecification as Bom_1_3,
)
from hoppr_cyclonedx_models.cyclonedx_1_4 import (
    CyclonedxSoftwareBillOfMaterialsStandard as Bom_1_4,
)
from hoppr_cyclonedx_models.cyclonedx_1_4 import Vulnerability
from rich.progress import Progress, SpinnerColumn, TextColumn
from security_commons.common.vulnerability_scanner import VulnerabilitySuper
from security_commons.common.vulnerability_combiner import combine_vulnerabilities


class CombinedScanner(VulnerabilitySuper):
    """A Vulnerability Scanner that combines results from all configured scanners"""

    scanners: List[VulnerabilitySuper] = []

    def set_scanners(self, scanners: Union[List[VulnerabilitySuper], List[str]]):
        """sets the scanners that should be used for vulnerability scanning.
        The argument can either be a list of scanner instances or a list of fully qualified strings to a scanner
         instance. For example ["vuln.gemnasium.gemnasium_scanner.GemnasiumScanner"]
        """
        for scanner in scanners:

            if isinstance(scanner, str):
                modname, _, clsname = scanner.rpartition(".")
                mod = importlib.import_module(modname)
                scanner = getattr(mod, clsname)()

            if scanner.should_activate():
                typer.echo(f"{scanner.__class__.__name__} is activated")
                self.scanners.append(scanner)

    def ___scan_concurrently(
        self, function
    ) -> dict[str, Optional[list[Vulnerability]]]:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            transient=True,
        ) as progress:
            results = []
            progress.add_task(description="Fetching vulnerabilities...", total=None)
            with concurrent.futures.ThreadPoolExecutor(max_workers=6) as executor:
                futures = {
                    executor.submit(function, scanner): scanner
                    for scanner in self.scanners
                }
            for future in concurrent.futures.as_completed(futures):
                scanner = futures[future]
                try:

                    scanner_results = future.result()
                    results.append(scanner_results)
                    # if not found:
                    #     failed.append(purl)
                except Exception as exc:  # pylint: disable=broad-except
                    print(f"{scanner.__class__.__name__} generated an exception: {exc}")
        return combine_vulnerabilities(list(results))

    def get_vulnerabilities_by_purl(
        self, purls: list[PackageURL]
    ) -> dict[str, Optional[list[Vulnerability]]]:
        """Get the vulnerabilities for a list of package URLS (purls)
        This function will return a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found
        """

        def submit_to_scanner_purl(scanner: VulnerabilitySuper):
            return scanner.get_vulnerabilities_by_purl(purls)

        return self.___scan_concurrently(submit_to_scanner_purl)

    def get_vulnerabilities_by_sbom(
        self, bom: [Union[Bom_1_4, Bom_1_3]]
    ) -> dict[str, Optional[list[Vulnerability]]]:
        """Parse a cyclone dx 1.4 compatible BOM and return a list of vulnerabilities "
        This function will return a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found
        """

        def submit_to_scanner(scanner: VulnerabilitySuper):
            return scanner.get_vulnerabilities_by_sbom(bom)

        return self.___scan_concurrently(submit_to_scanner)



"""A vulnerability scanner that locates vulnerabilities in Sonotypes' OSS Index. """
# This file is part of hoppr-cop
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Copyright (c) 2022 Lockheed Martin Corporation
import os
from typing import Optional, List

from cvss import CVSS3, CVSS2
from cvss.exceptions import CVSS3MalformedError
from hoppr_cyclonedx_models.cyclonedx_1_4 import Vulnerability, Rating, Severity, Tool
from packageurl import PackageURL
from requests.auth import HTTPBasicAuth
from security_commons.common.utils import (
    get_vulnerability_source,
    get_advisories_from_urls,
)
from security_commons.common.vulnerability_scanner import VulnerabilitySuper

from hopprcop.ossindex.api.model import OssIndexComponent
from hopprcop.ossindex.api.model import Vulnerability as OssVulnerability
from hopprcop.ossindex.api.ossindex import OssIndex


class OSSIndexScanner(VulnerabilitySuper):
    """A vulnerability scanner that locates vulnerabilities in Sonotypes' OSS Index."""

    required_environment_variables = ["OSS_INDEX_TOKEN", "OSS_INDEX_USER"]
    api = OssIndex()
    api.osthentication = HTTPBasicAuth(
        os.getenv("OSS_INDEX_TOKEN"), os.getenv("OSS_INDEX_USER")
    )

    supported_types = [
        "npm",
        "maven",
        "pypi",
        "gem",
        "golang",
        "nuget",
        "rpm",
        "connan",
    ]

    def get_vulnerabilities_by_purl(
        self, purls: list[PackageURL]
    ) -> dict[str, Optional[list[Vulnerability]]]:
        """Get the vulnerabilities for a list of package URLS (purls)
        This function will return a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found
        """
        purls = list(filter(lambda x: x.type in self.supported_types, purls))
        cleaned_purl_map = {}

        def remove_qualifiers(pkg_url: PackageURL) -> PackageURL:
            before_cleaning = pkg_url.to_string()
            pkg_url.qualifiers.clear()
            if pkg_url.type == "rpm":
                pkg_url = PackageURL(
                    name=pkg_url.name, type=pkg_url.type, version=pkg_url.version
                )
            cleaned_purl_map[pkg_url.to_string()] = before_cleaning
            return pkg_url

        purls = list(map(remove_qualifiers, purls))

        results: List[OssIndexComponent] = self.api.get_component_report(packages=purls)
        enhanced_results = {}

        for result in results:
            purl = result.coordinates
            enhanced_results[cleaned_purl_map[purl]] = []
            for vulnerability in result.vulnerabilities:
                enhanced_results[cleaned_purl_map[purl]].append(
                    self.__convert_to_cyclone_dx(vulnerability)
                )
        return enhanced_results

    @staticmethod
    def __convert_to_cyclone_dx(vulnerability: OssVulnerability) -> Vulnerability:
        """convert an OSS Index vulnerability to cyclone dx."""
        vuln_id = (
            vulnerability.cve
            if vulnerability.cve is not None
            else vulnerability.display_name
        )
        cwes = [int(vulnerability.cwe.replace("CWE-", ""))]

        cyclone_vuln = Vulnerability(
            id=vuln_id,
            description=vulnerability.description,
            cwes=cwes,
            source=get_vulnerability_source(vuln_id),
        )
        cyclone_vuln.ratings = []
        if vulnerability.cvss_vector is not None:
            try:
                cvss = CVSS3(vulnerability.cvss_vector)
                cyclone_vuln.ratings.append(
                    Rating(
                        score=cvss.base_score,
                        severity=Severity[cvss.severities()[0].lower()],
                        method="CVSSv3",
                        vector=cvss.vector,
                    )
                )
            except CVSS3MalformedError:
                cvss = CVSS2(vulnerability.cvss_vector)
                cyclone_vuln.ratings.append(
                    Rating(
                        score=cvss.base_score,
                        severity=Severity[cvss.severities()[0].lower()],
                        method="CVSSv2",
                        vector=cvss.vector,
                    )
                )
        cyclone_vuln.advisories = get_advisories_from_urls(
            list(vulnerability.external_references)
        )
        cyclone_vuln.tools = [Tool(vendor="Sonotype", name="OSS-Index")]

        return cyclone_vuln



# pylint: disable=missing-class-docstring
"""This file contains auto generated pydantic models for the grype json output"""
# This file is part of hoppr-cop
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Copyright (c) 2022 Lockheed Martin Corporation
from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field


class Fix(BaseModel):
    versions: List
    state: str


class Metrics(BaseModel):
    base_score: float = Field(..., alias="baseScore")
    exploitability_score: float = Field(..., alias="exploitabilityScore")
    impact_score: float = Field(..., alias="impactScore")


class Cvs(BaseModel):
    version: str
    vector: str
    metrics: Metrics
    vendor_metadata: Dict[str, Any] = Field(..., alias="vendorMetadata")


class Vulnerability(BaseModel):
    id: str
    data_source: str = Field(..., alias="dataSource")
    namespace: str
    severity: str
    urls: List[str]
    description: Optional[str]
    cvss: List[Cvs]
    fix: Fix
    advisories: List


class RelatedVulnerability(BaseModel):
    id: str
    data_source: str = Field(..., alias="dataSource")
    namespace: str
    severity: str
    urls: List[str]
    description: str
    cvss: List[Cvs]


class SearchedBy(BaseModel):
    language: Optional[str]
    namespace: Optional[str]


class Found(BaseModel):
    version_constraint: str = Field(..., alias="versionConstraint")


class MatchDetail(BaseModel):
    type: str
    matcher: str
    searched_by: SearchedBy = Field(..., alias="searchedBy")
    found: Found


class Artifact(BaseModel):
    name: str
    version: str
    type: str
    locations: List
    language: str
    licenses: List
    cpes: List
    purl: str
    upstreams: List


class Match(BaseModel):
    vulnerability: Vulnerability
    related_vulnerabilities: List[RelatedVulnerability] = Field(
        ..., alias="relatedVulnerabilities"
    )
    match_details: List[MatchDetail] = Field(..., alias="matchDetails")
    artifact: Artifact


class Source(BaseModel):
    type: str
    # target: [Optional[str]]


class Distro(BaseModel):
    name: str
    version: str
    id_like: Any = Field(..., alias="idLike")


class Search(BaseModel):
    scope: str
    unindexed_archives: bool = Field(..., alias="unindexed-archives")
    indexed_archives: bool = Field(..., alias="indexed-archives")


class Db(BaseModel):
    cache_dir: str = Field(..., alias="cache-dir")
    update_url: str = Field(..., alias="update-url")
    ca_cert: str = Field(..., alias="ca-cert")
    auto_update: bool = Field(..., alias="auto-update")
    validate_by_hash_on_start: bool = Field(..., alias="validate-by-hash-on-start")
    validate_age: bool = Field(..., alias="validate-age")
    max_allowed_built_age: int = Field(..., alias="max-allowed-built-age")


class Maven(BaseModel):
    search_upstream_by_sha1: bool = Field(..., alias="searchUpstreamBySha1")
    base_url: str = Field(..., alias="baseUrl")


class ExternalSources(BaseModel):
    enable: bool
    maven: Maven


class Dev(BaseModel):
    profile_cpu: bool = Field(..., alias="profile-cpu")
    profile_mem: bool = Field(..., alias="profile-mem")


class Registry(BaseModel):
    insecure_skip_tls_verify: bool = Field(..., alias="insecure-skip-tls-verify")
    insecure_use_http: bool = Field(..., alias="insecure-use-http")
    auth: List


class Log(BaseModel):
    structured: bool
    level: str
    file: str


class Attestation(BaseModel):
    public_key: str = Field(..., alias="public-key")
    skip_verification: bool = Field(..., alias="skip-verification")


class Configuration(BaseModel):
    config_path: str = Field(..., alias="configPath")
    output: str
    file: str
    distro: str
    add_cpes_if_none: bool = Field(..., alias="add-cpes-if-none")
    output_template_file: str = Field(..., alias="output-template-file")
    quiet: bool
    check_for_app_update: bool = Field(..., alias="check-for-app-update")
    only_fixed: bool = Field(..., alias="only-fixed")
    only_notfixed: bool = Field(..., alias="only-notfixed")
    platform: str
    search: Search
    ignore: Any
    exclude: List
    db: Db
    external_sources: ExternalSources = Field(..., alias="externalSources")
    dev: Dev
    fail_on_severity: str = Field(..., alias="fail-on-severity")
    registry: Registry
    log: Log
    attestation: Attestation


class Db1(BaseModel):
    built: str
    schema_version: int = Field(..., alias="schemaVersion")
    location: str
    checksum: str
    error: Any


class Descriptor(BaseModel):
    name: str
    version: str
    configuration: Configuration
    db: Db1


class GrypeResult(BaseModel):
    matches: List[Match] = []
    source: Optional[Source] = None
    distro: Optional[Distro] = None
    descriptor: Optional[Descriptor] = None



""""Interacts with the grype cli to scan a sbom"""
# This file is part of hoppr-cop
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Copyright (c) 2022 Lockheed Martin Corporation
# pylint: disable=duplicate-code
import json
from subprocess import PIPE, Popen
from typing import Optional, Union

from cvss import CVSS3, CVSS2
from hoppr_cyclonedx_models.cyclonedx_1_3 import (
    CyclonedxSoftwareBillOfMaterialSpecification as Bom_1_3,
)
from hoppr_cyclonedx_models.cyclonedx_1_4 import (
    CyclonedxSoftwareBillOfMaterialsStandard as Bom_1_4,
)
from hoppr_cyclonedx_models.cyclonedx_1_4 import Vulnerability, Rating, Severity, Tool
from packageurl import PackageURL
from security_commons.common.utils import (
    get_vulnerability_source,
    get_advisories_from_urls,
    get_references_from_ids,
    build_bom_from_purls,
)
from security_commons.common.vulnerability_scanner import VulnerabilitySuper

from hopprcop.grype.models import Match, GrypeResult


class GrypeScanner(VulnerabilitySuper):
    """This scanner utilizes the anchore grype command line to gather vulnerabilities"""

    required_tools_on_path = ["grype"]

    def __init__(self):
        super()

    def get_vulnerabilities_by_purl(
        self, purls: list[PackageURL]
    ) -> dict[str, Optional[list[Vulnerability]]]:
        """Get the vulnerabilities for a list of package URLS (purls)
        This function will return a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found
        """
        bom = build_bom_from_purls(purls)
        return self.get_vulnerabilities_by_sbom(bom)

    def get_vulnerabilities_by_sbom(
        self, bom: [Union[Bom_1_4, Bom_1_3]]
    ) -> dict[str, Optional[list[Vulnerability]]]:
        """Parse a cyclone dx 1.4 compatible BOM and return a list of vulnerabilities "
        This function will return a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found
        """
        with Popen(
            ["grype", "--output", "json"], stdout=PIPE, stdin=PIPE, stderr=PIPE
        ) as process:
            stdout_data = process.communicate(input=(bytes(bom.json(), "utf-8")))[0]
            result = GrypeResult(**json.loads(stdout_data))
            results = {}
        for component in bom.components:
            if component.purl is not None:
                results[component.purl] = []

        for match in list(result.matches):
            purl = PackageURL.from_string(match.artifact.purl)
            if not (purl.type == "npm" and purl.namespace == "@types"):
                results[match.artifact.purl].append(self.__convert_to_cyclone_dx(match))
        return results

    @staticmethod
    def __convert_to_cyclone_dx(match: Match) -> Vulnerability:
        """Converts a match to a vulnerability"""

        related = (
            match.vulnerability
            if len(match.related_vulnerabilities) == 0
            else match.related_vulnerabilities[0]
        )
        for related_vuln in match.related_vulnerabilities:
            if related_vuln.id.startswith("CVE"):
                related = related_vuln
                break

        cyclone_vuln = Vulnerability(
            id=related.id,
            description=related.description,
            ratings=[],
            recommendation=(
                f"State: {match.vulnerability.fix.state} | "
                f"Fix Versions: {','.join(match.vulnerability.fix.versions)}"
            ),
            source=get_vulnerability_source(related.id),
        )

        ids = [match.vulnerability.id] + list(
            map(lambda x: x.id, match.related_vulnerabilities)
        )

        cyclone_vuln.source.url = related.data_source
        cyclone_vuln.advisories = get_advisories_from_urls(related.urls)
        cyclone_vuln.references = get_references_from_ids(ids, cyclone_vuln.id)
        cvss_scores = (
            match.vulnerability.cvss
            if len(match.vulnerability.cvss) > 0
            else related.cvss
        )
        for cvss in cvss_scores:
            if cvss.version.startswith("3"):
                cvss3 = CVSS3(cvss.vector)
                method = "CVSSv31" if cvss.version == "3.1" else "CVSSv3"
                cyclone_vuln.ratings.append(
                    Rating(
                        score=cvss3.base_score,
                        severity=Severity[cvss3.severities()[0].lower()],
                        method=method,
                        vector=cvss.vector,
                    )
                )
            elif cvss.version.startswith("2"):
                cvss2 = CVSS2(cvss.vector)
                cyclone_vuln.ratings.append(
                    Rating(
                        score=cvss2.base_score,
                        severity=Severity[cvss2.severities()[0].lower()],
                        method="CVSSv2",
                        vector=cvss.vector,
                    )
                )
        if len(cyclone_vuln.ratings) == 0 and match.vulnerability.severity is not None:
            cyclone_vuln.ratings.append(
                Rating(
                    severity=Severity[match.vulnerability.severity.lower()],
                    method="other",
                )
            )
        cyclone_vuln.tools = [Tool(vendor="Anchore", name="Grype")]
        return cyclone_vuln



"""A Vulnerability Scanner for Gitlab's Gemnasiumm Database """
# This file is part of hoppr-cop
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Copyright (c) 2022 Lockheed Martin Corporation
import os
import pkgutil
import shutil
import stat
import subprocess
import tempfile
import time
import zipfile
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse
import requests
import typer
import yaml
from cvss import CVSS2, CVSS3
from hoppr_cyclonedx_models.cyclonedx_1_4 import Vulnerability, Rating, Severity, Tool
from packageurl import PackageURL

from security_commons.common.utils import (
    get_vulnerability_source,
    get_advisories_from_urls,
    get_references_from_ids,
)
from security_commons.common.vulnerability_scanner import VulnerabilitySuper
from hopprcop.gemnasium.models import GemnasiumVulnerability


class GemnasiumScanner(VulnerabilitySuper):
    """A Vulnerability Scanner for Gitlab's Gemnasiumm Database"""

    supported_formats = ["npm", "maven", "pypi", "gem", "golang", "connan"]
    database_path = Path(tempfile.gettempdir()) / "gemnasium"
    # url = "https://gitlab.com/gitlab-org/security-products/gemnasium-db/-/archive/master/gemnasium-db-master.zip"
    url = os.getenv(
        "GEMNASIUM_DATABASE_ZIP",
        "https://gitlab.com/gitlab-org/advisories-community/-/archive/main/advisories-community-main.zip",
    )

    semver_path = "/usr/local/bin/semver"
    required_tools_on_path = ["ruby"]

    def __init__(self):
        if not Path(self.semver_path).exists():
            self.__extract_semver_to_local()
        self.__download_and_extract_database()

    def __extract_semver_to_local(self):
        """If the ruby semver command isn't installed then extract from this package"""
        data = pkgutil.get_data(__name__, "semver").decode("utf-8")
        self.semver_path = Path(tempfile.gettempdir()) / "semver"
        with open(self.semver_path, "w", encoding="UTF-8") as file:
            file.write(data)
            file.close()
        os.chmod(
            self.semver_path,
            stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH,
        )

    def __download_and_extract_database(self):
        """Downloads the gymnasium database"""
        url = urlparse(self.url)

        path_to_zip_file = Path(tempfile.gettempdir()) / os.path.basename(url.path)

        def do_download_and_unpack():
            typer.echo(f"Updating Gemnasium database to {self.database_path}")
            request = requests.get(self.url, allow_redirects=True)
            with open(path_to_zip_file, "wb") as file:
                file.write(request.content)
                file.close()

            with zipfile.ZipFile(path_to_zip_file, "r") as zip_ref:
                zip_ref.extractall(self.database_path)

        if not path_to_zip_file.exists():
            do_download_and_unpack()
        else:
            file_time = os.path.getmtime(path_to_zip_file)
            # Check against 24 hours
            older_than_one_day = (time.time() - file_time) / 3600 > 24 * 1
            if older_than_one_day:
                shutil.rmtree(self.database_path / path_to_zip_file.stem)
                do_download_and_unpack()
        self.database_path = self.database_path / path_to_zip_file.stem

    def __is_affected_range(self, repository_format, version, affected_range) -> bool:
        """
        Checks if the version matches the affected range based on package manager specific semantic versioning.
        :param repository_format:
        :param version:
        :param affected_range:
        :return:
        """
        try:

            output = subprocess.run(
                [
                    self.semver_path,
                    "check_version",
                    repository_format,
                    version,
                    affected_range,
                ],
                capture_output=True,
                text=True,
                check=False,
                # TODO this command suddenly started throwing errors, not sure what changed but it needs investigated.
                # It looks like  a spell checker package changed
            )
            return "matches" in str(output)
        except Exception as err:  # pylint: disable=broad-except
            print(f"Failed to check version for: {repository_format} {version} {err}")
            return False

    def get_vulnerabilities_by_purl(
        self, purls: list[PackageURL]
    ) -> dict[str, Optional[list[Vulnerability]]]:
        """Get the vulnerabilities for a list of package URLS (purls)
        This function will return a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found
        """
        vulnerabilities_by_purl = {}
        for purl in purls:
            vulnerabilities_by_purl[purl.to_string()] = []
            path = self.__get_path(purl)
            if path.exists():
                for filename in os.listdir(path):
                    try:
                        if (path / filename).is_file():
                            with open(path / filename, "r", encoding="UTF-8") as file:
                                data = yaml.full_load(file)
                                file.close()
                                vuln = GemnasiumVulnerability(**data)

                                if self.__is_affected_range(
                                    purl.type, purl.version, vuln.affected_range
                                ):
                                    vulnerability = self.__convert_to_cyclone_dx(vuln)
                                    if len(vulnerability.ratings) > 0:
                                        vulnerabilities_by_purl[
                                            purl.to_string()
                                        ].append(vulnerability)
                    except:  # pylint: disable=bare-except
                        print(f"failed to parse gemnasium file for {purl}")

        return vulnerabilities_by_purl

    @staticmethod
    def __convert_to_cyclone_dx(vuln: GemnasiumVulnerability) -> Vulnerability:
        # pylint: disable=duplicate-code
        """Converts a gemnasium vulnerabity to a vulnerability"""
        res = list(filter(lambda x: "cve-" in x.lower(), vuln.identifiers))
        vuln_id = res[0] if len(res) > 1 else vuln.identifiers[0]
        cwes = []
        if vuln.cwe_ids is not None:
            cwes = list(map(lambda x: int(x.replace("CWE-", "")), vuln.cwe_ids))
        cyclone_vuln = Vulnerability(
            id=vuln_id,
            recommendation=vuln.solution,
            cwes=cwes,
            description=vuln.description,
            ratings=[],
            source=get_vulnerability_source(vuln_id),
        )
        cyclone_vuln.advisories = get_advisories_from_urls(vuln.urls)
        cyclone_vuln.references = get_references_from_ids(
            vuln.identifiers, cyclone_vuln.id
        )
        if vuln.cvss_v3 is not None:
            cvss = CVSS3(vuln.cvss_v3)
            cyclone_vuln.ratings.append(
                Rating(
                    score=cvss.base_score,
                    severity=Severity[cvss.severities()[0].lower()],
                    method="CVSSv3",
                    vector=str(cvss.clean_vector()),
                )
            )
        elif vuln.cvss_v2 is not None:
            cvss = CVSS2(vuln.cvss_v2)
            cyclone_vuln.ratings.append(
                Rating(
                    score=cvss.base_score,
                    severity=Severity[cvss.severities()[0].lower()],
                    method="CVSSv2",
                    vector=cvss.clean_vector(),
                )
            )
        cyclone_vuln.tools = [Tool(vendor="Gitlab", name="Gemnasium")]
        return cyclone_vuln

    def __get_path(self, purl: PackageURL):
        """build a path to the gemnasium path"""
        repo_format = purl.type
        if repo_format == "npm":
            if purl.namespace != "" and purl.namespace is not None:
                path_slug = f"npm/{purl.namespace}/{purl.name}"
            else:
                path_slug = f"npm/{purl.name}"
        elif repo_format == "maven":
            path_slug = f"maven/{purl.namespace}/{purl.name}"
        elif repo_format == "golang":
            path_slug = f"go/{purl.namespace}/{purl.name}"
        else:
            path_slug = f"{repo_format}/{purl.name}"

        return self.database_path / Path(path_slug)



"""Models fo Gitlab Gemnasium Vulnerabilities """
# This file is part of hoppr-cop
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Copyright (c) 2022 Lockheed Martin Corporation
from typing import List, Optional

from pydantic import BaseModel


class GemnasiumVulnerability(BaseModel):
    """Models fo Gitlab Gemnasium Vulnerabilities"""

    identifiers: List[str]
    title: str
    description: str
    pubdate: str
    solution: Optional[str] = ""
    affected_range: str
    affected_versions: str
    not_impacted: Optional[str] = None
    urls: List[str] = []
    cwe_ids: Optional[List[str]]
    cvss_v2: Optional[str] = None
    cvss_v3: Optional[str] = None



""""Interacts with the trivy cli to scan an sbom"""
# This file is part of hoppr-cop
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Copyright (c) 2022 Lockheed Martin Corporation
import json
import tempfile
from subprocess import PIPE, Popen
from typing import Optional

from hoppr_cyclonedx_models.cyclonedx_1_4 import (
    CyclonedxSoftwareBillOfMaterialsStandard as Bom_1_4,
)
from hoppr_cyclonedx_models.cyclonedx_1_4 import Vulnerability, Tool
from packageurl import PackageURL

from security_commons.common.utils import (
    build_bom_from_purls,
)
from security_commons.common.vulnerability_scanner import VulnerabilitySuper


class TrivyScanner(VulnerabilitySuper):
    """ "Interacts with the trivy cli to scan an sbom"""

    required_tools_on_path = ["trivy"]
    supported_types = ["npm", "maven", "pypi", "gem", "golang", "nuget", "connan"]

    def get_vulnerabilities_by_purl(
        self, purls: list[PackageURL]
    ) -> dict[str, Optional[list[Vulnerability]]]:
        """Get the vulnerabilities for a list of package URLS (purls)
        This function will return a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found
        """
        results = {}
        for purl in purls:
            results[purl.to_string()] = []

        purls = list(filter(lambda x: x.type in self.supported_types, purls))
        if len(purls) > 0:
            bom = build_bom_from_purls(purls)

            with tempfile.NamedTemporaryFile(mode="w") as bom_file:
                bom_file.write(bom.json())

                with Popen(
                    ["trivy", "sbom", "--format", "cyclonedx", str(bom_file.name)],
                    stdout=PIPE,
                    stdin=PIPE,
                    stderr=PIPE,
                ) as process:
                    stdout_data = process.communicate(input=b"")[0]
                    bom_file.close()
            bom_dict = json.loads(stdout_data)
            bom_dict["metadata"]["component"]["type"] = "application"
            bom_dict["metadata"]["component"]["name"] = "generated"
            trivy_result = Bom_1_4(**bom_dict)

            for vuln in trivy_result.vulnerabilities:
                for affects in vuln.affects:
                    _, _, purl = str(affects.ref).partition("#")
                    affects.ref.__root__ = purl.strip("'")
                    if vuln.ratings is not None:
                        results[str(affects.ref.__root__)].append(vuln)
                vuln.tools = [Tool(vendor="Aquasec", name="Trivy")]

        return results



# pylint: skip-file
"""models for interacting with oss index api"""
#
# Copyright 2022-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Iterable, Optional, Set

# See https://github.com/package-url/packageurl-python/issues/65
from packageurl import PackageURL  # type: ignore

# pylint: disable=invalid-name


class Vulnerability:
    """
    Model class that represents a Vulnerability as received back from OSS Index.

    """

    def __init__(
        self,
        *,
        id_: str,
        display_name: str,
        title: str,
        description: str,
        cvss_score: Optional[float] = None,
        cvss_vector: Optional[str] = None,
        cve: Optional[str] = None,
        cwe: Optional[str] = None,
        version_ranges: Optional[Iterable[str]] = None,
        reference: str,
        external_references: Optional[Iterable[str]] = None,
    ):
        self.id = id_
        self.display_name = display_name
        self.title = title
        self.description = description
        self.cvss_score = cvss_score
        self.cvss_vector = cvss_vector
        self.cve = cve
        self.cwe = cwe
        self.version_ranges = set(version_ranges or [])
        self.reference = reference
        self.external_references = set(external_references or [])

    @property
    def id(self) -> str:
        """
        OSS Index's unique UUID for this Vulnerability.

        Returns:
             `str`
        """
        return self._id

    @id.setter
    def id(self, id_: str) -> None:
        self._id = id_

    @property
    def display_name(self) -> str:
        """
        displayName returned by OSS Index

        Returns:
            `str`
        """
        return self._display_name

    @display_name.setter
    def display_name(self, display_name: str) -> None:
        self._display_name = display_name

    @property
    def title(self) -> str:
        """
        title returned by OSS Index

        Returns:
            `str`
        """
        return self._title

    @title.setter
    def title(self, title: str) -> None:
        self._title = title

    @property
    def description(self) -> str:
        """
        description returned by OSS Index.

        Returns:
             `str`
        """
        return self._description

    @description.setter
    def description(self, description: str) -> None:
        self._description = description

    @property
    def cvss_score(self) -> Optional[float]:
        """
        CVSS Score returned from OSS Index.

        Returns:
             `float` if set else `None`
        """
        return self._cvss_score

    @cvss_score.setter
    def cvss_score(self, cvss_score: Optional[float]) -> None:
        self._cvss_score = cvss_score

    @property
    def cvss_vector(self) -> Optional[str]:
        """
        CVSS Vector returned from OSS Index

        Returns:
            `str` if set else `None`
        """
        return self._cvss_vector

    @cvss_vector.setter
    def cvss_vector(self, cvss_vector: Optional[str]) -> None:
        self._cvss_vector = cvss_vector

    @property
    def cwe(self) -> Optional[str]:
        """
        CWE returned from OSS Index.

        .. note:
            This is a string of the format CWE-nnn or an empty string

        Returns:
             `str` if set else `None`
        """
        return self._cwe

    @cwe.setter
    def cwe(self, cwe: Optional[str]) -> None:
        self._cwe = cwe

    @property
    def cve(self) -> Optional[str]:
        """
        CVE returned from OSS Index.

        Returns:
             `str` if set else `None`
        """
        return self._cve

    @cve.setter
    def cve(self, cve: Optional[str]) -> None:
        self._cve = cve

    @property
    def reference(self) -> str:
        """
        Reference URL to OSS Index for this Vulnerability.

        Returns:
            `str`
        """
        return self._reference

    @reference.setter
    def reference(self, reference: str) -> None:
        self._reference = reference

    @property
    def version_ranges(self) -> Set[str]:
        """
        Range of versions which are impacted by this Vulnerability.

        Returns:
            Set of `str`
        """
        return self._version_ranges

    @version_ranges.setter
    def version_ranges(self, version_ranges: Iterable[str]) -> None:
        self._version_ranges = set(version_ranges)

    @property
    def external_references(self) -> Set[str]:
        """
        List of external references that provide additional information about the vulnerability.

        Returns:
            Set of `str`
        """
        return self._external_references

    @external_references.setter
    def external_references(self, external_references: Iterable[str]) -> None:
        self._external_references = set(external_references)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Vulnerability):
            return hash(other) == hash(self)
        return False

    def __hash__(self) -> int:
        return hash(
            (
                self.id,
                self.display_name,
                self.title,
                self.description,
                self.cvss_score,
                self.cvss_vector,
                self.cve,
                self.cwe,
                tuple(self.version_ranges),
                self.reference,
                tuple(self.external_references),
            )
        )

    def __repr__(self) -> str:
        return f"<Vulnerability id={self.id}, name={self.display_name}, cvss_score={self.cvss_score:.1f}>"


class OssIndexComponent:
    """
    Model class that represents a Component Report as received back from OSS Index.

    """

    def __init__(
        self,
        *,
        coordinates: str,
        description: Optional[str] = None,
        reference: str,
        vulnerabilities: Optional[Iterable[Vulnerability]] = None,
    ) -> None:
        self.coordinates = coordinates
        self.description = description
        self.reference = reference
        self.vulnerabilities = set(vulnerabilities or [])

    @property
    def coordinates(self) -> str:
        """
        PackageURL formatted coordinates of this Component.

        Returns:
             `str`
        """
        return self._coordinates

    @coordinates.setter
    def coordinates(self, coordinates: str) -> None:
        self._coordinates = coordinates

    @property
    def description(self) -> Optional[str]:
        """
        Description of the Component from OSS Index.

        Returns:
            `str` if set else `None`
        """
        return self._description

    @description.setter
    def description(self, description: Optional[str]) -> None:
        self._description = description

    @property
    def reference(self) -> str:
        """
        URL to this Component on OSS Index.

        Returns:
             `str`
        """
        return self._reference

    @reference.setter
    def reference(self, reference: str) -> None:
        self._reference = reference

    @property
    def vulnerabilities(self) -> Set[Vulnerability]:
        """
        Known vulnerabilities that relate to this Component.

        Returns:
             Set of `Vulnerability`
        """
        return self._vulnerabilities

    @vulnerabilities.setter
    def vulnerabilities(self, vulnerabilities: Iterable[Vulnerability]) -> None:
        self._vulnerabilities = set(vulnerabilities)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, OssIndexComponent):
            return hash(other) == hash(self)
        return False

    def __hash__(self) -> int:
        return hash(
            (
                self.coordinates,
                self.description,
                self.reference,
                tuple(self.vulnerabilities),
            )
        )

    def __repr__(self) -> str:
        return f"<OssIndexComponent coordinates={self.coordinates}>"

    def get_package_url(self) -> PackageURL:
        """
        Get a PURL representation of this components coordinates.

        Returns:
            `PackageURL`
        """
        return PackageURL.from_string(purl=self.coordinates)

    def get_max_cvss_score(self) -> float:
        """
        Get the maximum CVSS Score across all Vulnerabilities known for this Component.

        Returns:
             `float`
        """
        max_cvss_score = 0.0
        if self.vulnerabilities:
            for v in self.vulnerabilities:
                max_cvss_score = OssIndexComponent._reduce_on_max_cvss_score(
                    v=v, current_max=max_cvss_score
                )
        return max_cvss_score

    @staticmethod
    def _reduce_on_max_cvss_score(v: Vulnerability, current_max: float) -> float:
        if v.cvss_score:
            if v.cvss_score > current_max:
                return v.cvss_score
        return current_max



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hopprcop',
 'hopprcop.combined',
 'hopprcop.gemnasium',
 'hopprcop.grype',
 'hopprcop.hoppr_plugin',
 'hopprcop.ossindex',
 'hopprcop.ossindex.api',
 'hopprcop.trivy']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'coverage>=6.5.0,<7.0.0',
 'cvss>=2.5,<3.0',
 'hoppr-cyclonedx-models>=0.2.1,<0.3.0',
 'hoppr-security-commons>=0.0.6,<0.0.7',
 'hoppr>=1.6.2,<2.0.0',
 'mkdocs-mermaid2-plugin>=0.6.0,<0.7.0',
 'mkdocs>=1.3.1,<2.0.0',
 'packageurl-python>=0.10.1,<0.11.0',
 'pytest==7.2.0',
 'rich>=12.5.1,<13.0.0',
 'tabulate>=0.9.0,<0.10.0',
 'tinydb>=4.7.0,<5.0.0',
 'typer>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['hoppr-cop = hopprcop.combined.cli:app']}

setup_kwargs = {
    'name': 'hoppr-cop',
    'version': '1.0.18',
    'description': '',
    'long_description': 'None',
    'author': 'kganger',
    'author_email': 'keith.e.ganger@lmco.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



# pylint: skip-file
#
# Copyright 2022-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import logging
import os
import sys
import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import requests
import yaml

# See https://github.com/package-url/packageurl-python/issues/65
from packageurl import PackageURL  # type: ignore
from tinydb import TinyDB, Query
from tinydb.table import Document

from .exception import AccessDeniedException
from .model import OssIndexComponent
from .serializer import json_decoder, OssIndexJsonEncoder

logger = logging.getLogger("ossindex")

if sys.version_info >= (3, 8):
    from importlib.metadata import version as meta_version
else:
    from importlib_metadata import version as meta_version

ossindex_lib_version: str = "TBC"
try:
    ossindex_lib_version = str(meta_version("ossindex-lib"))  # type: ignore[no-untyped-call]
except Exception:
    ossindex_lib_version = "DEVELOPMENT"


class OssIndex:
    DEFAULT_CONFIG_FILE = ".oss-index.config"

    _caching_enabled: bool = False
    _cache_directory: str = ".ossindex"
    _cache_ttl_in_hours: int = 12

    _oss_index_api_version: str = "v3"
    _oss_index_host: str = "https://ossindex.sonatype.org"
    _oss_max_coordinates_per_request: int = 128
    oss_index_authentication: Optional[requests.auth.HTTPBasicAuth] = None

    def __init__(
        self, *, enable_cache: bool = True, cache_location: Optional[str] = None
    ) -> None:
        self._caching_enabled = enable_cache
        if self._caching_enabled:
            logger.info("OssIndex caching is ENABLED")
            self._setup_cache(cache_location=cache_location)
        self._attempt_config_load()

    def has_ossindex_authentication(self) -> bool:
        return self.oss_index_authentication is not None

    def _attempt_config_load(self) -> None:
        try:
            config_filename: str = os.path.join(
                os.path.expanduser("~"), OssIndex.DEFAULT_CONFIG_FILE
            )
            with open(config_filename, "r") as ossindex_confg_f:
                ossindex_config = yaml.safe_load(ossindex_confg_f.read())
                self.oss_index_authentication = requests.auth.HTTPBasicAuth(
                    ossindex_config["username"], ossindex_config["password"]
                )
        except FileNotFoundError:
            pass

    def get_component_report(
        self, packages: List[PackageURL]
    ) -> List[OssIndexComponent]:
        logger.debug(
            "A total of {} Packages to be queried against OSS Index".format(
                len(packages)
            )
        )
        return self._get_results(packages=packages)

    def purge_local_cache(self) -> None:
        if self._caching_enabled:
            logger.info("Truncating local cache database as requested")
            with self._get_cache_db() as db:
                db.truncate()
                logger.info("Local OSS Index cache has been purged")

    def _chunk_packages_for_oss_index(
        self, packages: List[PackageURL]
    ) -> List[List[PackageURL]]:
        """
        Splits up the list of packages into lists that are of a size consumable by OSS Index
        APIs.

        :param packages: List[PackageURL]
        :return: List[List[PackageURL]]
        """
        return list(
            [
                packages[i : i + self._oss_max_coordinates_per_request]
                for i in range(0, len(packages), self._oss_max_coordinates_per_request)
            ]
        )

    def _get_api_url(self, api_uri: str) -> str:
        return "{}/api/{}/{}".format(
            self._oss_index_host, self._oss_index_api_version, api_uri
        )

    def _get_cached_results(
        self, packages: List[PackageURL]
    ) -> Tuple[List[PackageURL], List[OssIndexComponent]]:
        """
        Takes a list of packages and returns two Lists:
            1. Packages without cached results
            2. Cached results for those packages where they exist

        :param packages: List[PackageURL]
        :return: (List[PackageURL], List[OssIndexComponent])
        """
        if not self._caching_enabled:
            # This should not be possible, but adding for developer safety
            return packages, []

        now = datetime.now()
        cached_results: List[OssIndexComponent] = []
        non_cached_packaged: List[PackageURL] = []
        with self._get_cache_db() as cache_db:
            for package in packages:
                logger.debug("   Checking in cache for {}".format(package.to_string()))
                cache_results: List[Document] = cache_db.search(
                    Query().coordinates == package.to_string()
                )
                if len(cache_results) == 0:
                    # Not cached
                    logger.debug("      Not in cache")
                    non_cached_packaged.append(package)
                elif (
                    datetime.strptime(
                        cache_results[0]["expiry"], "%Y-%m-%dT%H:%M:%S.%f"
                    )
                    < now
                ):
                    logger.debug("      Cached, but result expired")
                    non_cached_packaged.append(package)
                else:
                    logger.debug("      Cached, loading from cache")
                    cached_results.append(
                        json.loads(
                            f'[{cache_results[0]["response"]}]',
                            object_hook=json_decoder,
                        ).pop()
                    )

        return non_cached_packaged, cached_results

    @staticmethod
    def _get_headers() -> Dict[str, str]:
        return {
            "Accept": "application/vnd.ossindex.component-report.v1+json",
            "Content-type": "application/vnd.ossindex.component-report-request.v1+json",
            "User-Agent": f"python-oss-index-lib@{ossindex_lib_version}",
        }

    def _get_results(self, packages: List[PackageURL]) -> List[OssIndexComponent]:
        results: List[OssIndexComponent] = list()

        # First get any cached results
        if self._caching_enabled:
            logger.debug("Checking local cache for any usable results...")
            packages, results = self._get_cached_results(packages=packages)
            logger.debug(
                "   {} cached results found leaving {} to ask OSS Index".format(
                    len(results), len(packages)
                )
            )

        # Second, chunk up packages for which we have no cached results and query OSS Index
        chunk: List[PackageURL]
        chunks = self._chunk_packages_for_oss_index(packages=packages)
        logger.debug(
            "Split {} packages into {} chunks for OSS requests".format(
                len(packages), len(chunks)
            )
        )
        for chunk in chunks:
            logger.debug("  Getting chunk results from OSS Index...")
            results = results + self._make_oss_index_component_report_call(
                packages=chunk
            )

        logger.debug("Total of {} results (including cached)".format(len(results)))
        return results

    def _make_oss_index_component_report_call(
        self, packages: List[PackageURL]
    ) -> List[OssIndexComponent]:
        response = requests.post(
            url=self._get_api_url("component-report"),
            headers=self._get_headers(),
            json={"coordinates": list(map(lambda p: str(p.to_string()), packages))},
            auth=self.oss_index_authentication,
        )

        if not response.status_code == 200:
            print(response.status_code)
            if response.status_code == 429:
                time.sleep(90)
            raise AccessDeniedException()

        results: List[OssIndexComponent] = []
        for oic in response.json(object_hook=json_decoder):
            results.append(oic)

        if self._caching_enabled:
            self._upsert_cache_with_oss_index_responses(oss_components=results)
        return results

    def _upsert_cache_with_oss_index_responses(
        self, oss_components: List[OssIndexComponent]
    ) -> None:
        if not self._caching_enabled:
            return

        now = datetime.now()
        cache_expiry = now + timedelta(hours=self._cache_ttl_in_hours)
        oc: OssIndexComponent
        with self._get_cache_db() as cache_db:
            for oc in oss_components:
                query: Query = Query()
                cache_query_result: List[Document] = cache_db.search(
                    query.coordinates == oc.coordinates
                )
                if len(cache_query_result) == 0:
                    # New component for caching
                    logger.debug(
                        "    Caching new Component results for {}".format(
                            oc.coordinates
                        )
                    )
                    cache_db.insert(
                        {
                            "coordinates": oc.coordinates,
                            "response": json.dumps(oc, cls=OssIndexJsonEncoder),
                            "expiry": cache_expiry.isoformat(),
                        }
                    )
                else:
                    # Update existing cache
                    logger.debug(
                        "    Might refresh cache for {}".format(oc.coordinates)
                    )
                    if now > datetime.strptime(
                        cache_query_result[0]["expiry"], "%Y-%m-%dT%H:%M:%S.%f"
                    ):
                        # Cache expired - update it!
                        logger.debug(
                            "        Cache expired for {} - UPDATING CACHE".format(
                                oc.coordinates
                            )
                        )
                        cache_db.update(
                            {
                                "response": json.dumps(oc, cls=OssIndexJsonEncoder),
                                "expiry": cache_expiry.isoformat(),
                            },
                            query.coordinates == oc.coordinates,
                        )
                    else:
                        logger.debug(
                            "    Cache is still valid for {} - not updating".format(
                                oc.coordinates
                            )
                        )

    def _get_cache_db(self) -> TinyDB:
        return TinyDB(os.path.join(self._cache_directory, "ossindex.json"))

    def _setup_cache(self, cache_location: Optional[str] = None) -> None:
        full_cache_path: str
        if not cache_location:
            full_cache_path = os.path.join(Path.home(), self._cache_directory)
        else:
            full_cache_path = os.path.join(cache_location, self._cache_directory)

        if not os.path.exists(full_cache_path):
            Path(full_cache_path).mkdir(parents=True, exist_ok=True)

        self._cache_directory = str(Path(full_cache_path))
